# Copyright (c) 2015 Ultimaker B.V.
# Uranium is released under the terms of the AGPLv3 or higher.

from UM.Tool import Tool
from UM.Job import Job
from UM.Logger import Logger
from UM.Event import Event, MouseEvent, KeyEvent
from UM.Application import Application
from UM.Message import Message
from UM.Scene.ToolHandle import ToolHandle
from UM.Scene.Selection import Selection
from UM.Operations.OperationStack import OperationStack

from UM.Math.Plane import Plane
from UM.Math.Vector import Vector
from UM.Math.Quaternion import Quaternion
from UM.Math.Float import Float

from UM.Operations.RotateOperation import RotateOperation
from UM.Operations.GroupedOperation import GroupedOperation
from UM.Operations.SetTransformOperation import SetTransformOperation
from UM.Operations.LayFlatOperation import LayFlatOperation

from UM.Qt.Bindings.ActiveToolProxy import ActiveToolProxy

from . import RotateToolHandle
from .GenceRotateOperation import GenceRotateOperation

from cura.SIProxies.NotificationsProxy import NotificationsProxy
import math
import time


##  Provides the tool to rotate meshes and groups
#
#   The tool exposes a ToolHint to show the rotation angle of the current operation

class GenceRotateTool(Tool):
    singleStep = 0.0174532925;

    def __init__(self):
        super().__init__()
        self._isViewVisible = False;
        self._handle = RotateToolHandle.RotateToolHandle()
        self._snap_rotation = True
        self._snap_angle = math.radians(15)

        self._angle = None
        self._angle_update_time = None
        self._inprogess = False
        self._rotationCounts = 0

        self._progress_message = None
        self._iterations = 0
        self._total_iterations = 0
        self._rotating = False
        self.setExposedProperties("ToolHint", "RotationSnap", "RotationSnapAngle", "RotationX", "RotationY",
                                  "RotationZ")
        self._saved_node_positions = []
        self._notifications = NotificationsProxy.createNotificationsProxy()
        self._operationStack =Application.getInstance().getOperationStack()


    def rotate(self, value, axis, vector = None):
        if (Selection.getAllSelectedObjects()):
            value %= 360
            
            node = Selection.getAllSelectedObjects()[0]

            node_rotation = {"x": node._rotationX, "y": node._rotationY, "z": node._rotationZ}

            old_quaternion = Quaternion().fromAngleAxis(math.radians(node_rotation["x"]), Vector(1, 0, 0))
            old_quaternion *= Quaternion().fromAngleAxis(math.radians(node_rotation["y"]), Vector(0, 1, 0))
            old_quaternion *= Quaternion().fromAngleAxis(math.radians(node_rotation["z"]), Vector(0, 0, 1))

            old_quaternion.invert()

            if (axis == "_rotationX"):
                node_rotation["x"] = value
            elif (axis == "_rotationY"):
                node_rotation["y"] = value
            elif (axis == "_rotationZ"):
                node_rotation["z"] = value

            new_quaternion = Quaternion().fromAngleAxis(math.radians(node_rotation["x"]), Vector(1, 0, 0))
            new_quaternion *= Quaternion().fromAngleAxis(math.radians(node_rotation["y"]), Vector(0, 1, 0))
            new_quaternion *= Quaternion().fromAngleAxis(math.radians(node_rotation["z"]), Vector(0, 0, 1))

            new_quaternion = old_quaternion * new_quaternion
            node._rotation = new_quaternion

            setattr(node, axis, value)

            self.propertyChanged.emit()

            GenceRotateOperation(node, node._rotation, tuple(node_rotation.values()), self.propertyChanged,
                                 rotate_around_point=node.getPosition(), mergeKey=axis).push()

            self.setNodeAtZeroLevel(node)

    def getRotation(self, axis):
        for node in Selection.getAllSelectedObjects():
            return getattr(node, axis, 0)
        return 0

    def setRotationX(self, value):
        self.rotate(value=value, axis="_rotationX", vector=Vector.Unit_X)

    def getRotationX(self):
        return self.getRotation(axis="_rotationX")

    def setRotationY(self, value):
        self.rotate(value=value, axis="_rotationY", vector=Vector.Unit_Y)

    def getRotationY(self):
        return self.getRotation(axis="_rotationY")

    def setRotationZ(self, value):
        self.rotate(value=value, axis="_rotationZ", vector=Vector.Unit_Z)

    def getRotationZ(self):
        return self.getRotation(axis="_rotationZ")

    ##  Handle mouse and keyboard events
    #
    #   \param event type(Event)

    def showView(self):
        self._notifications.getCallback("onShowRotateView")()

    def event(self, event):
        super().event(event)
        self._operationStack._merge_operations = False
        if event.type ==  Event.ToolActivateEvent:

            for node in Selection.getAllSelectedObjects():
                y = getattr(node, "_rotationY", 0)
                x = getattr(node, "_rotationX",0)
                z = getattr(node, "_rotationZ",0)
                setattr(node, "_rotationX", x)
                setattr(node, "_rotationY", y)
                setattr(node, "_rotationZ", z)

            self.propertyChanged.emit()
            self._notifications.getCallback("onShowRotateView")()
            return True
        else :
            if event.type == Event.ToolDeactivateEvent:
                self.propertyChanged.emit()
                self._operationStack._merge_operations=False
                return True

        if event.type == Event.MousePressEvent and self._controller.getToolsEnabled():

            # Start a rotate operation
            if MouseEvent.LeftButton not in event.buttons:
                return False

            id = self._selection_pass.getIdAtPosition(event.x, event.y)
            if not id:
                return False

            self.setLockedAxis(ToolHandle.YAxis)

            handle_position = self._handle.getWorldPosition()

            self._saved_node_positions = []
            for node in Selection.getAllSelectedObjects():
                self._saved_node_positions.append((node, node.getPosition()))
                value = getattr(node, "_rotationY", 0)
                setattr(node, "_rotationY", value)

            self.setDragPlane(Plane(Vector(0, 1, 0), handle_position.y))

            self.setDragStart(event.x, event.y)
            self._rotating = False
            self._angle = 0

        if event.type == Event.MouseMoveEvent:

            if not self.getDragPlane():
                return False

            if not self.getDragStart():
                self.setDragStart(event.x, event.y)

            if not self._rotating:
                self._rotating = True
                # self.operationStarted.emit(self)

            handle_position = self._handle.getWorldPosition()
            drag_start = (self.getDragStart() - handle_position).normalized()
            drag_position = self.getDragPosition(event.x, event.y)
            if not drag_position:
                return
            drag_end = (drag_position - handle_position).normalized()


            try:
                angle = math.acos(drag_start.dot(drag_end))

            except ValueError:
                angle = 0

            direction = 1 if Vector.Unit_Y.dot(drag_start.cross(drag_end)) > 0 else -1
            # We want whole values only
            degree = math.degrees(angle)
            degree = math.floor(degree)
            angle = math.radians(degree)

            new_time = time.monotonic()

            # Only values bigger then 1 degree, and not to often
            if degree >= 1 and not self._inprogess and (
                not self._angle_update_time or new_time - self._angle_update_time > 0.1):

                self._inprogess = True

                self._angle_update_time = new_time
                self._angle += direction * angle
                self.propertyChanged.emit()

                if (self._saved_node_positions):
                    node, position = self._saved_node_positions[0]
                    current_rotationY = node._rotationY
                    result_angle = current_rotationY + direction*degree
                    self.setRotationY(result_angle)

                self.setDragStart(event.x, event.y)
                self._inprogess = False

        if event.type == Event.MouseReleaseEvent:

            # Finish a rotate operation
            if self.getDragPlane():
                self.setDragPlane(None)
                self.setLockedAxis(None)
                self._angle = None

                if self._rotating:
                    pass
                    #self.operationStopped.emit(self)
                return True

    ##  Return a formatted angle of the current rotate operation
    #
    #   \return type(String) fully formatted string showing the angle by which the mesh(es) are rotated
    def getToolHint(self):
        return "%d°" % round(math.degrees(self._angle)) if self._angle else None

    ##  Get the state of the "snap rotation to N-degree increments" option
    #
    #   \return type(Boolean)
    def getRotationSnap(self):
        return self._snap_rotation

    ##  Set the state of the "snap rotation to N-degree increments" option
    #
    #   \param snap type(Boolean)
    def setRotationSnap(self, snap):
        if snap != self._snap_rotation:
            self._snap_rotation = snap
            self.propertyChanged.emit()

    ##  Get the number of degrees used in the "snap rotation to N-degree increments" option
    #
    #   \return type(Number)
    def getRotationSnapAngle(self):
        return self._snap_angle

    ##  Set the number of degrees used in the "snap rotation to N-degree increments" option
    #
    #   \param snap type(Number)
    def setRotationSnapAngle(self, angle):
        if angle != self._snap_angle:
            self._snap_angle = angle
            self.propertyChanged.emit()

    ##  Reset the orientation of the mesh(es) to their original orientation(s)
    def resetRotation(self):

        if (Selection.getAllSelectedObjects ()):

            node = Selection.getAllSelectedObjects () [0]

            node_rotation = {"x": node._rotationX, "y": node._rotationY, "z": node._rotationZ}

            quaternion  = Quaternion().fromAngleAxis (math.radians (node_rotation["x"]), Vector (1, 0, 0))
            quaternion *= Quaternion().fromAngleAxis (math.radians (node_rotation["y"]), Vector (0, 1, 0))
            quaternion *= Quaternion().fromAngleAxis (math.radians (node_rotation["z"]), Vector (0, 0, 1))

            quaternion.invert ()

            node._rotation = quaternion

            setattr (node, '_rotationX', 0)
            setattr (node, '_rotationY', 0)
            setattr (node, '_rotationZ', 0)

            self.propertyChanged.emit ()

            GenceRotateOperation (node, node._rotation, tuple(node_rotation.values()), self.propertyChanged,
                                 rotate_around_point = node.getPosition (), mergeKey = None).push ()

            self.setNodeAtZeroLevel (node)

    ##  Initialise and start a LayFlatOperation
    #
    #   Note: The LayFlat functionality is mostly used for 3d printing and should probably be moved into the Cura project
    def layFlat(self):
        # self.operationStarted.emit(self)
        self._progress_message = Message("Laying object flat on buildplate...", lifetime=0, dismissable=False)
        self._progress_message.setProgress(0)

        self._iterations = 0
        self._total_iterations = 0
        for selected_object in Selection.getAllSelectedObjects():
            if not selected_object.callDecoration("isGroup"):
                self._total_iterations += len(selected_object.getMeshDataTransformed().getVertices()) * 2
            else:
                for child in selected_object.getChildren():
                    self._total_iterations += len(child.getMeshDataTransformed().getVertices()) * 2

        self._progress_message.show()

        operations = Selection.applyOperation(LayFlatOperation)
        for op in operations:
            op.progress.connect(self._layFlatProgress)

        job = LayFlatJob(operations)
        job.finished.connect(self._layFlatFinished)
        job.start()

    ##  Called while performing the LayFlatOperation so progress can be shown
    #
    #   Note that the LayFlatOperation rate-limits these callbacks to prevent the UI from being flooded with property change notifications,
    #   \param iterations type(int) number of iterations performed since the last callback
    def _layFlatProgress(self, iterations):
        self._iterations += iterations
        self._progress_message.setProgress(100 * self._iterations / self._total_iterations)

    ##  Called when the LayFlatJob is done running all of its LayFlatOperations
    #
    #   \param job type(LayFlatJob)
    def _layFlatFinished(self, job):
        if self._progress_message:
            self._progress_message.hide()
            self._progress_message = None

        # self.operationStopped.emit(self)


##  A LayFlatJob bundles multiple LayFlatOperations for multiple selected objects
#
#   The job is executed on its own thread, processing each operation in order, so it does not lock up the GUI.
class LayFlatJob(Job):
    def __init__(self, operations):
        super().__init__()

        self._operations = operations

    def run(self):
        for op in self._operations:
            op.process()
