from UM.Tool import Tool
from UM.Event import Event
from UM.Application import Application
from UM.Scene.Selection import Selection

from cura.SIProxies.NotificationsProxy import NotificationsProxy
from .SimpleSplitMesh import SimpleSplitMesh
from .AdvancedSplitMesh import AdvancedSplitMesh
from .Split import Axis
from .ModelSplitter import ModelSplitter
from . import GenceSpliterPlane

import math

class GenceSplitterTool(Tool):
    class CutFunctionType:
        Simple = 1
        Divide = 2
        Advanced = 3

    def __init__(self):
        super().__init__()

        # Simple & advanced
        self._cutType = None
        self._axis = None
        self._cutValue = 0

        # Only Advanced
        self._edge = 1           # mm
        self._angle = 45         # degrees
        self._offset = 0      # mm
        self._length = 5         # mm
        self._invert = False     # bool

        self._min = 10
        self._max = 50

        self._cutPlane = None

        self._controller = Application.getInstance().getController()
        self._scene = Application.getInstance().getController().getScene()
        self._render = Application.getInstance().getRenderer()
        self._notifications = NotificationsProxy.createNotificationsProxy()

        self.setExposedProperties("CutMin", "CutMax")

    def event(self, event):
        super().event(event)

        if event.type == Event.ToolActivateEvent:
            self._cutType = GenceSplitterTool.CutFunctionType.Simple
            self._axis = Axis.X
            self._notifications.getCallback("onShowSplitTool")()
            self._updateCutPlane(False)

        if event.type == Event.ToolDeactivateEvent:
            self._getPlane().hideSplitPlane()

    def getCutMin(self):
        return self._min

    def getCutMax(self):
        return self._max

    def setCutMin(self, value):
        self._min = value
        self.propertyChanged.emit()

    def setCutMax(self, value):
        self._max = value
        self.propertyChanged.emit()

    def setCutType(self, type):
        if type == "Divide":
            self._cutType = GenceSplitterTool.CutFunctionType.Divide
        elif type == "Advanced":
            self._cutType = GenceSplitterTool.CutFunctionType.Advanced
        else:
            self._cutType = GenceSplitterTool.CutFunctionType.Simple
        self._updateCutPlane()

    def setAxis(self, axis):
        if axis == "Y":
            self._axis = Axis.Y
        elif axis == "Z":
            self._axis = Axis.Z
        else:
            self._axis = Axis.X

        self._updateCutPlane()

    def setValue(self, value: float):
        self._cutValue = float(value)
        self._getPlane().setPlanePosition(self._axis, self._cutValue)

    def setEdge(self, value: float):
        self._edge = float(round(value*4)/4)

    def setAngle(self, value: int):
        self._angle = int(value)

    def setOffset(self, value: float):
        self._offset = float(round(value*4)/4)

    def setLength(self, value: float):
        self._length = float(round(value*4)/4)

    def setInvert(self, value: bool):
        self._invert = bool(value)

    def runFunction(self):
        if self._cutType == GenceSplitterTool.CutFunctionType.Divide:
            ModelSplitter()
        else:
            if self._axis == Axis.Y:
                axis = Axis.Z
            elif self._axis == Axis.Z:
                axis = Axis.Y
            else:
                axis = Axis.X

            if self._cutType == GenceSplitterTool.CutFunctionType.Simple:
                SimpleSplitMesh(axis, self._cutValue)
            if self._cutType == GenceSplitterTool.CutFunctionType.Advanced:
                AdvancedSplitMesh(axis, self._cutValue, self._edge, self._angle, self._offset, self._length, self._invert)

    def _updateCutPlane(self, notify=True):
        if self._cutType == GenceSplitterTool.CutFunctionType.Divide:
            self._getPlane().hideSplitPlane()
            return

        bb = Selection.getBoundingBox()

        if self._axis == Axis.X:
            self.setCutMin(bb.left.item())
            self.setCutMax(bb.right.item())
        if self._axis == Axis.Y:
            self.setCutMin(bb.back.item())
            self.setCutMax(bb.front.item())
        if self._axis == Axis.Z:
            self.setCutMin(bb.bottom.item())
            self.setCutMax(bb.top.item())

        if notify:
            self._notifications.getCallback("onUpdateSplitTool")()
        else:
            self._cutValue = int((math.floor(self.getCutMin()) + math.floor(self.getCutMax()))/2)

        self._getPlane().showSplitPlane(self._axis, bb, bb.center)

    def _getPlane(self):
        if self._cutPlane is None:
            self._cutPlane = GenceSpliterPlane.GenceSpliterPlane()

        return self._cutPlane
