import QtQuick 2.4
import QtQuick.Layouts 1.2

import "../singletons"
import "../tweaks"
import "../common"

import Gence 1.0 as Gence
import Cura 1.0 as Cura
import UM 1.2 as UM

Rectangle {
    id: button
    property alias button: button
    property alias deactivateButton: deactivateIndicator.visible
    property alias helpText: buttonTweaks.helpText
    property alias mouseArea: mouseArea
    property color buttonBackgroundColor: Colors.buttonBackground
    property bool buttonOpened: Printer.isDetailsOpened
    property double buttonHeight: buttonOpened ? Globals.buttonHeight + columnLayout.implicitHeight : Globals.buttonHeight
    property double iconWidth: buttonOpened ? 0 : 1
    property variant profileData: Printer.profileData
    property string material0Name: profileData["material0"]
    property string material1Name: profileData["material1"]
    property string producer0Name: profileData["producer0"]
    property string producer1Name: profileData["producer1"]
    property bool customProfile: Printer.inCustom
    property bool multipleExtrudersProfile: machineExtruderCount.properties.value > 1
    property bool supportMaterialTemperature: (customProfile && multipleExtrudersProfile) || (material0Name != "" && material1Name != "")
    property bool detailsVisible: buttonOpened && !heightAnimation.running
    property bool buttonTextVisible: !buttonOpened && !heightAnimation.running
    Layout.preferredHeight: buttonHeight
    Layout.fillWidth: true
    color: buttonBackgroundColor

    Behavior on buttonHeight { NumberAnimation { id: heightAnimation; duration: 50; alwaysRunToEnd: true } }
    Behavior on iconWidth { NumberAnimation { duration: heightAnimation.duration; alwaysRunToEnd: true } }

    MouseArea {
        id: mouseArea
        visible: !deactivateIndicator.visible
        anchors.fill: parent
    }

    RowLayout{
        anchors.fill: parent
        spacing: 0

        ColumnLayout {
            anchors.fill: parent
            spacing: 0

            Rectangle {
                id: textBackground
                Layout.preferredHeight: Globals.buttonHeaderHeight 
                Layout.fillWidth: true
                color: Colors.buttonTextBackground

                Text {
                    id: buttonNameText
                    x: parent.height - Globals.fontPixelSize
                    anchors.verticalCenter: parent.verticalCenter
                    font.pixelSize: Globals.fontPixelSize
                    font.capitalization: Font.AllUppercase
                    font.family: Fonts.boxed

                    color: Colors.buttonText
                    text: "printing profile details"
                }
            }

            Rectangle {
                id: decorator
                Layout.fillWidth: true
                Layout.preferredHeight: Globals.decoratorHeight
                color: Colors.buttonDecorator
            }

            Rectangle {
                id: buttonBackground
                Layout.fillHeight: true
                Layout.fillWidth: true
                color: buttonBackgroundColor

                ColumnLayout {
                    id: columnLayout
                    //TODO: instead of using visible, it's rather ugly, but there was some unexplained property loop on setting visibility.
                    scale: detailsVisible ? 1.0 : 0.0
                    spacing: 0

                    anchors {
                        fill: parent
                        leftMargin: (button.width - Globals.buttonHeight) * 0.05
                        topMargin: anchors.leftMargin
                        rightMargin: anchors.leftMargin
                    }

                    ProfileDetailText {
                        visible: material0Name != "" && !customProfile
                        label: "Model material type"
                        value: Printer.profileData["material0"]
                    }

                    ProfileDetailText {
                        visible: producer0Name != "" && !customProfile
                        label: "Producer"
                        value: Printer.profileData["producer0"]
                    }

                    ProfileDetailText {
                        visible: customProfile
                        label: "Custom profile loaded"
                    }

                    ProfileDetailText {
                        visible: customProfile
                        label: "Profile name"
                        value: Gence.CustomSettingsActions.profileName
                    }

                    Spacer {
                        Layout.fillHeight: false
                        Layout.preferredHeight: parent.anchors.leftMargin - parent.spacing * 4
                    }

                    ProfileDetailText {
                        visible: material1Name != "" && !customProfile && multipleExtrudersProfile
                        label: "Support material type"
                        value: Printer.profileData["material1"]
                    }

                    ProfileDetailText {
                        visible: producer1Name != "" && !customProfile && multipleExtrudersProfile
                        label: "Producer"
                        value: Printer.profileData["producer1"]
                    }

                    Spacer {
                        visible: material1Name != "" && !customProfile && multipleExtrudersProfile
                        Layout.fillHeight: false
                        Layout.preferredHeight: parent.anchors.leftMargin - parent.spacing * 4
                    }

                    ProfileSettingText {
                        label: "Layer height"
                        settingKey: "layer_height"
                        machineSetting: true
                    }

                    ProfileSettingText {
                        label: "Infill density"
                        settingKey: "infill_sparse_density"
                    }

                    ProfileSettingText {
                        label: "Outer walls"
                        settingKey: "wall_line_count"
                    }

                    ProfileSettingText {
                        label: "Top layers"
                        settingKey: "top_layers"
                    }

                    ProfileSettingText {
                        label: "Bottom layers"
                        settingKey: "bottom_layers"
                    }

                    ProfileSettingText {
                        label: "Heatbed adhesion type"
                        settingKey: "adhesion_type"
                        machineSetting: true
                    }

                    ProfileSettingText {
                        label: "Model material temperature"
                        settingKey: "material_print_temperature"
                    }

                    ProfileSettingText {
                        visible: supportMaterialTemperature
                        label: "Support material temperature"
                        settingKey: "material_print_temperature"
                        extruder: 1
                    }

                    Spacer {
                        Layout.fillHeight: false
                        Layout.preferredHeight: parent.anchors.topMargin
                    }
                }

                Text {
                    id: buttonText
                    visible: buttonTextVisible
                    x: parent.width * 0.1
                    text: "Press for details"
                    anchors.verticalCenter: parent.verticalCenter
                    font.pixelSize: Globals.fontPixelSize
                    font.family: Fonts.boxed
                    color: Colors.buttonOptionText
                    width: parent.width * 0.85
                    elide: Text.ElideMiddle
                }

                //Corner triangle
                Image {
                    id: corner
                    visible: buttonTextVisible
                    height: Globals.buttonTriangleSize
                    y: (parent.height * 0.93) - corner.height
                    x: parent.height - y - height

                    mirror: true
                    fillMode: Image.PreserveAspectFit
                    source: "../images/corner.svg"
                }
            }
        }

        Item {
            Layout.preferredHeight: parent.height
            Layout.preferredWidth: Globals.buttonHeight * iconWidth
            Layout.maximumWidth: Globals.buttonHeight * iconWidth

            Image {
                id: buttonIcon
                anchors.top: parent.top
                anchors.topMargin: (parent.width - width) / 2
                anchors.horizontalCenter: parent.horizontalCenter
                height: Globals.rightMenuButtonIconSize * iconWidth
                mipmap: true

                fillMode: Image.PreserveAspectFit

                source: "../images/info_icon.svg"
            }
        }
    }

    Rectangle {
        id: deactivateIndicator
        anchors.fill: parent
        visible: false
        color: Colors.buttonDeactivator
    }

    BottomShadow {}

    ButtonTweaks{
        id: buttonTweaks
        hoverIndicatorEnabled: false
        helpEnabled: !deactivateIndicator.visible
    }

    UM.SettingPropertyProvider
    {
        id: machineExtruderCount

        containerStackId: Cura.MachineManager.activeMachineId
        key: "machine_extruder_count"
        watchedProperties: [ "value" ]
        storeIndex: 0
    }
}
