# Copyright (c) 2015 Ultimaker B.V.
# Uranium is released under the terms of the AGPLv3 or higher.

from PyQt5.QtCore import pyqtProperty, QObject, Qt, QCoreApplication, pyqtSignal, pyqtSlot, QMetaObject, QRectF,QTimer
from PyQt5.QtGui import QColor
from PyQt5.QtQuick import QQuickWindow, QQuickItem


from UM.Math.Vector import Vector
from UM.Math.Matrix import Matrix
from UM.Qt.QtMouseDevice import QtMouseDevice
from UM.Qt.QtKeyDevice import QtKeyDevice
from UM.Application import Application
from UM.Preferences import Preferences
from threading import Thread, Lock

##  QQuickWindow subclass that provides the main window.
class MainWindow(QQuickWindow):
    def __init__(self, parent = None):
        super(MainWindow, self).__init__(parent)

        self._background_color = QColor(204, 204, 204, 255)



        self._mutex = Lock()
        self._update_mutex=Lock()


        self.setClearBeforeRendering(False)
        self.beforeRendering.connect(self._render, type=Qt.DirectConnection)

        self._mouse_device = QtMouseDevice(self)
        self._mouse_device.setPluginId("qt_mouse")
        self._key_device = QtKeyDevice()
        self._key_device.setPluginId("qt_key")
        self._previous_focus = None

        self._app = QCoreApplication.instance()
        self._app.getController().addInputDevice(self._mouse_device)
        self._app.getController().addInputDevice(self._key_device)
        self._app.getController().getScene().sceneChanged.connect(self._onSceneChanged)
        self._preferences = Preferences.getInstance()

        self._preferences.addPreference("general/window_width", 1280)
        self._preferences.addPreference("general/window_height", 720)
        self._preferences.addPreference("general/window_left", 50)
        self._preferences.addPreference("general/window_top", 50)
        self._preferences.addPreference("general/window_state", Qt.WindowNoState)

        # self._change_timer = QTimer()
        # self._change_timer.setInterval(250)
        # self._change_timer.setSingleShot(True)
        # self._change_timer.timeout.connect(self._recalculateViewport)
        # self._initialized = False
        # self._size = None


        # Restore window geometry
        self.setWidth(int(self._preferences.getValue("general/window_width")))
        self.setHeight(int(self._preferences.getValue("general/window_height")))
        self.setPosition(int(self._preferences.getValue("general/window_left")), int(self._preferences.getValue("general/window_top")))
        # Make sure restored geometry is not outside the currently available screens
        screen_found = False
        for s in range(0, self._app.desktop().screenCount()):
            if self.geometry().intersects(self._app.desktop().availableGeometry(s)):
                screen_found = True
                break
        if not screen_found:
            self.setPosition(50,50)

        self.setWindowState(int(self._preferences.getValue("general/window_state")))
        self._mouse_x = 0
        self._mouse_y = 0

        self._viewport_rect = QRectF(0, 0, 1.0, 1.0)

        Application.getInstance().setMainWindow(self)
        self._fullscreen = False

    @pyqtSlot()
    def toggleFullscreen(self):
        if self._fullscreen:
            self.setVisibility(QQuickWindow.Windowed) # Switch back to windowed
        else:
            self.setVisibility(QQuickWindow.FullScreen) # Go to fullscreen
        self._fullscreen = not self._fullscreen

    def getBackgroundColor(self):
        return self._background_color

    def setBackgroundColor(self, color):
        self._background_color = color
        self._app.getRenderer().setBackgroundColor(color)

    backgroundColor = pyqtProperty(QColor, fget=getBackgroundColor, fset=setBackgroundColor)

    mousePositionChanged = pyqtSignal()

    @pyqtProperty(int, notify = mousePositionChanged)
    def mouseX(self):
        return self._mouse_x

    @pyqtProperty(int, notify = mousePositionChanged)
    def mouseY(self):
        return self._mouse_y

    def setViewportRect(self, rect):
        if rect != self._viewport_rect:
            self._viewport_rect = rect
            self._updateViewportGeometry(self.width() * self.devicePixelRatio(), self.height() * self.devicePixelRatio())
            self.viewportRectChanged.emit()

    viewportRectChanged = pyqtSignal()

    @pyqtProperty(QRectF, fset = setViewportRect, notify = viewportRectChanged)
    def viewportRect(self):
        return self._viewport_rect

#   Warning! Never reimplemented this as a QExposeEvent can cause a deadlock with QSGThreadedRender due to both trying
#   to claim the Python GIL.
#   def event(self, event):

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if event.isAccepted():
            return

        if self.activeFocusItem() != None and self.activeFocusItem() != self._previous_focus:
            self.activeFocusItem().setFocus(False)

        self._previous_focus = self.activeFocusItem()
        self._mouse_device.handleEvent(event)

    def mouseMoveEvent(self, event):
        self._mouse_x = event.x()
        self._mouse_y = event.y()
        self.mousePositionChanged.emit()

        super().mouseMoveEvent(event)
        if event.isAccepted():
            return

        self._mouse_device.handleEvent(event)

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        if event.isAccepted():
            return
        self._mouse_device.handleEvent(event)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if event.isAccepted():
         return

        self._key_device.handleEvent(event)
        

    def keyReleaseEvent(self, event):
        super().keyReleaseEvent(event)
        if event.isAccepted():
            return

        self._key_device.handleEvent(event)

    def wheelEvent(self, event):
        super().wheelEvent(event)
        if event.isAccepted():
            return

        self._mouse_device.handleEvent(event)

    def moveEvent(self, event):
        QMetaObject.invokeMethod(self, "_onWindowGeometryChanged", Qt.QueuedConnection);

    def resizeEvent(self, event):

        super().resizeEvent(event)
        import copy
        self._size = copy.deepcopy(event.size())
        self._recalculateViewport()

        # 26th July, 2017 : Timer was switched off, because it probably wasn't using

        # if not self._initialized:
        #     self._initialized = True
        #     self._recalculateViewport()
        #
        # else:
        #     self._change_timer.start()

        # if camera and  not camera.isPerspective():
        # self._change_timer.start()


    def _recalculateViewport(self):

            self._mutex.acquire(False)
            camera = self._app.getController().getScene().getActiveCamera();

            win_w = self._size.width() * self.devicePixelRatio()
            win_h = self._size.height() * self.devicePixelRatio()

            QMetaObject.invokeMethod(self, "_onWindowGeometryChanged", Qt.QueuedConnection)

            if camera and not camera.isPerspective():

                self._updateViewportGeometry(win_w, win_h, False)

            else:
                self._updateViewportGeometry(win_w, win_h)

            tool = Application.getInstance().getController().getTool("CameraTool")
            global_container_stack = Application.getInstance().getGlobalContainerStack()

            if global_container_stack:
                width = global_container_stack.getProperty("machine_width", "value")
                if tool:
                    tool.setToTheSameDistanceinOrth(width)

                self._mutex.release()

                QMetaObject.invokeMethod(self, "_onWindowGeometryChanged", Qt.QueuedConnection)
            else:
                return

    def hideEvent(self, event):
        Application.getInstance().windowClosed()

    def _render(self):
        renderer = self._app.getRenderer()
        view = self._app.getController().getActiveView()

        renderer.beginRendering()
        view.beginRendering()
        renderer.render()
        view.endRendering()
        renderer.endRendering()

    def _onSceneChanged(self, object):
        self.update()

    @pyqtSlot()
    def _onWindowGeometryChanged(self):
        if self.windowState() == Qt.WindowNoState:
            self._preferences.setValue("general/window_width", self.width())
            self._preferences.setValue("general/window_height", self.height())
            self._preferences.setValue("general/window_left", self.x())
            self._preferences.setValue("general/window_top", self.y())
            self._preferences.setValue("general/window_state", Qt.WindowNoState)
        elif self.windowState() == Qt.WindowMaximized:
            self._preferences.setValue("general/window_state", Qt.WindowMaximized)


    def provideEnlargedOrthoView(self,zoom:float=0):

        camera = self._app.getController().getScene().getActiveCamera();

        if camera and not camera.isPerspective():
            renderer = self._app.getRenderer();
            size = renderer.getWindowSize()
            width,height = renderer.getWindowSize();
            view_w = int(width) * self._viewport_rect.width()
            view_h = int(height) * self._viewport_rect.height()

            camera.setViewportSize(view_w,view_h)
            camera.setWindowSize(width,height)

            proj= Matrix()
            divider = zoom+1*(view_w/200)
            proj.setOrtho(-view_w / divider, view_w / divider, -view_h / divider, view_h / divider, -5000, 5000)
            camera.setProjectionMatrix(proj)

            self._app.getRenderer().setViewportSize(view_w, view_h)
            self._app.getRenderer().setWindowSize(width, height)


    def updateViewportGeometry(self):
        windowSize=self.getWindowSize()
        self._updateViewportGeometry(windowSize[0],windowSize[1])

    def getWindowSize(self):
        return self._app.getRenderer().getWindowSize()

    def getOrthoProjection(self,zoom):
        windowSize = self._app.getRenderer().getWindowSize()
        viewDimension = self.calculateViewDimension(windowSize[0],windowSize[1])
        return self.calculateProjectionForOrtho(zoom,viewDimension[0],viewDimension[1])

    def calculateProjectionForOrtho(self,zoom,width,height):
        projectionMatrix = Matrix()

        value = zoom
        newWidth = width + zoom

        if(newWidth <0):
            if(zoom!=0):
                # sign = zoom/abs(zoom)
                # value = ( abs(zoom) % abs(width) )* sign * -1
                camera = self._app.getController().getScene().getActiveCamera()
                difference = abs(zoom) - width
                zoom = camera.getZoom() + difference*2
                camera.setZoom(zoom)
                newWidth = width + camera.getZoom()

        # newWidth = width + value
        # if(newWidth<0):
        #     newWidth=width


        newHeight = (newWidth/width) * height
        projectionMatrix.setOrtho(-newWidth,newWidth,-newHeight,newHeight,-10000,10000)

        return projectionMatrix

    def calculateViewDimension(self,width,height):
        # Ugly hack to set different view dimension due to politics of resizing menu (in qml)
        hackedViewportWidth = self._viewport_rect.width()
        if width/16 > height/9: 
            # Calculate right menu size in pixels
            rightMenuPixelWidth = int(height)*(16/9)*0.15
            # Calculate ratio of width without menu to whole width
            hackedViewportWidth = 1 - (rightMenuPixelWidth/int(width))

        view_w = int(width) * hackedViewportWidth
        view_h = int(height) * self._viewport_rect.height()

        return view_w, view_h

    def _updateViewportGeometry(self, width, height,withSet=True):
        # self._update_mutex.acquire()
        dimension = self.calculateViewDimension(width,height)
        view_w = dimension[0]
        view_h = dimension[1]

        for camera in self._app.getController().getScene().getAllCameras():
            proj = Matrix()
            proj.setPerspective(30, view_w / view_h, 1, 500)
            camera.setPerspectiveMatrix(proj)
            if camera.isPerspective():
                camera.setProjectionMatrix(proj)
            else:
                if withSet:
                    projectionMatrix = self.calculateProjectionForOrtho(camera.getZoom(),view_w,view_h)
                    camera.setProjectionMatrix(projectionMatrix)

            # if withSet:
            camera.setViewportSize(view_w, view_h)
            camera.setWindowSize(width, height)


        # if withSet:
        self._app.getRenderer().setViewportSize(view_w, view_h)
        self._app.getRenderer().setWindowSize(width, height)

        # self._update_mutex.release()
        