# Copyright (c) 2017 vabi vabi vabi
# Uranium is released under the terms of the AGPLv3 or higher.
# This class is also released under the terms of the AGPLv3 or higher

import math

from UM.Resources import Resources
from UM.Application import Application

from UM.Math.Color import Color

from UM.Scene.ToolHandle import ToolHandle
from UM.Scene.Iterator.DepthFirstIterator import DepthFirstIterator

from UM.View.RenderPass import RenderPass
from UM.View.RenderBatch import RenderBatch
from UM.View.GL.OpenGL import OpenGL

from PIL import Image
import PIL.ImageOps

##  A RenderPass subclass responsible for rendering shapes needed by Autoarranger
#
#   This pass performs rendering to array of bytes where 0 is pixel where no part
#   of object is rendered and 1 is pixel where object was rendered
class AutoarrangePass(RenderPass):
    def __init__(self, width, height):
        super().__init__("autoarrange", width, height, -999)

        self._shader = OpenGL.getInstance().createShaderProgram(Resources.getPath(Resources.Shaders, "autoarrange.shader"))
        self._gl = OpenGL.getInstance().getBindingsObject()
        self._scene = Application.getInstance().getController().getScene()

        self._renderer = Application.getInstance().getRenderer()

        self._output = None

        self._viewport_width = width
        self._viewport_height = height

    ##  Perform the actual rendering.
    def render(self):
        models_data = list()

        self._gl.glViewport(0, 0, self._viewport_width, self._viewport_height)
        for node in DepthFirstIterator(self._scene.getRoot()):
            if node.isSelectable() and node.getMeshData():
                batch = RenderBatch(self._shader)
                batch.addItem(transformation = node.getWorldTransformation(),
                              mesh = node.getMeshData())
                self.bind()

                camera = self._scene._findCamera("autoarranger_cam")
                batch.render(camera)
                self._gl.glColorMask(self._gl.GL_TRUE, self._gl.GL_TRUE, self._gl.GL_TRUE, self._gl.GL_FALSE)
                self._gl.glDisable(self._gl.GL_DEPTH_TEST)

                self.release()

                output = self.getOutput()
                model_image = PIL.ImageOps.invert(Image.fromqimage(output).convert('RGB'))
                model_image = model_image.convert('P', colors=1)
                image_box = model_image.getbbox()

                if not image_box:
                    # if nothing was rendered means that model is too small or too thin - create 1x1 pixel
                    # bitmap to represent it (it will be scaled to model's size by function in AutoarrangeTool)
                    # advisable - rewrite whole autoarranger with reasons for that sad tweaks in mind :(
                    model_image = Image.new("P", (1, 1), 255)

                if image_box:
                    model_image = model_image.crop(image_box)

                model_size = node._aabb._max - node._aabb._min
                model_significant_size = (math.ceil(model_size.x), math.ceil(model_size.z))
                models_data.append((id(node), model_image, model_significant_size))
                
        if len(models_data) != 0:
            models_data.sort(key=lambda modelField: modelField[2][0] * modelField[2][1], reverse = True)
            return models_data
        else:
            return None
